#
# TIConvert v0.1
# By Michael Lee
# Released Febuary 24 2011
#


import sys
import time
import argparse

from AxeTokensDict import *


### Defining metadata: ###
exe_name = "TIconvert.py"
program_name = "TI Tokenizer/Detokenizer"
program_version = "0.1"
release_date = "Feb 24 2011"
author_name = "Michael Lee"
python_version = "3.2"
axe_version = "0.5.0"

help_string = "{0} v{1}.  Released {2}.  Written by {3} using Python version {4}.  " \
              "Converts to Axe v{5}.  To view help, type 'TIconvert.py --help'"
help_string = help_string.format(program_name, 
                                 program_version, 
                                 release_date, 
                                 author_name, 
                                 python_version,
                                 axe_version)

### Defining data: ###
d_VariableTypes = {0:"a real number",
                   1:"a list",
                   2:"a matrix",
                   3:"an equation",
                   4:"a string",
                   5:"a program",
                   6:"a protected program",
                   7:"a picture",
                   8:"a GDB",
                   9:"unknown",
                   10:"an unknown equation",
                   11:"a 'new equation'",
                   12:"a complex number",
                   13:"a complex list",
                   14:"undefined",
                   15:"a window variable",
                   16:"a ZSto variable",
                   17:"a table range variable",
                   18:"a LCD variable",
                   19:"a backup",
                   20:"an application (app)",
                   21:"an appvar",
                   22:"a temporary program",
                   23:"a group"}

r_AllowedTypes = [5, 6, 21, 22]

hexstring = "0123456789ABCDEF"

signature = "2A2A54493833462A1A0A00"

is_logging_enabled = False


### Defining functions ###
def Write(message, override = 0):
    """This function is identical to print(X).

    To increase compatibility between Python 2.X and 3.X, this
    function ensures that 'print' is used only once for easy replacement.
    If logging is enabled, this also copies the text to the log, and
    if verbose is disabled, won't print anything to the console."""
    if options.verbose or (override == 2):
        print("\n" + message)
    if is_logging_enabled:
        try:
            log_file.write(message + "\n")
        except:
            pass
    return

def NumHexToStr(number):
    """Converts a decimal number into a hexadecimal string.

    >>> NumHexToStr(12)
    '0C' """
    temp1 = number // 16
    temp2 = number % 16

    result = hexstring[temp1] + hexstring[temp2]
    return result

def ReturnByte(offset, buffer):
    """ Reads a single-byte token from memory (string)

    Returns the integer equivalent"""
    temp1 = hexstring.index(buffer[offset * 2])
    temp2 = hexstring.index(buffer[offset * 2 + 1])
    return temp1 * 16 + temp2

def ReturnTwoBytes(offset, buffer):
    """ Reads a two-byte token from memory (string)

    Returns an integer equivalent"""
    return ReturnByte(offset + 1, buffer) * 256 + ReturnByte(offset, buffer)

def DoesFileExist(path):
    try:
        file = open(path, "a")
    except:
        Write("ERROR: " + path + " does not appear to exist - did you type it in correctly?")
        raise
        QuitProgram(error = 2)
    file.close()
    return

def QuitProgram(message="Terminating program...", error=0):
    """Terminates the program.

    Also tries closing any opened files."""
    Write(message)
    Write("Trying to close files...")
    try:
        input_file.close()
    except:
        Write("Input file already closed.")

    try:
        output_file.close()
    except:
        Write("Output file already closed.")

    try:
        log_file.close()
    except:
        pass
    
    sys.exit(error)
    return


### Defining the command line parser ###
parser = argparse.ArgumentParser(description = help_string,
                                 add_help = True)
parser.add_argument("input",
                    action = "store",
                    nargs = "?", 
                    help = "usage: TIconvert.py 'C:\Calc\etc\THEGAME.8xp'\n" \
                           "The path of the 8xp or text file you wish to convert.  " \
                           "This argument is mandatory.")
parser.add_argument("-p", "--path",
                    action = "store",
                    nargs = "?",
                    help = "usage: TIconvert.py -p 'C:\Calc\\random\'\n" \
                           "Changes the directory where the resulting file will be " \
                           "written to.  Defaults to the directory where the original " \
                           "is stored.",
                    dest = "target_path")
parser.add_argument("-l", "--log",
                    action = "store",
                    nargs = "?",
                    const = "!!!", 
                    help = "usage: TIconvert.py -l 'C:\Calc\misc'\n" \
                           "If you chose to use this argument, this program creates a " \
                           "log file at the given directory.  If the directory is not " \
                           "given, it creates the log file where the current program is " \
                           "stored.  This option is provided for debugging purposes only.")
parser.add_argument("-c", "--colon",
                    action = "store_true",
                    default = False,
                    help = "usage: TIconvert.py -c\n" \
                           "Use this if you want your text files to start with the colon at " \
                           "the start of each line.")
parser.add_argument("-v", "--verbose",
                    action = "store_true",
                    default = False,
                    help = "usage: TIconvert.py -v\n" \
                           "Prints loads of info to the console.  Somewhat useful for " \
                           "debugging, although enabling logging would be a better idea.  "\
                           "Enable this if you want visual feedback that something has happened.")
parser.add_argument("-t", "--tabs",
                    action = "store",
                    nargs = "?", 
                    default = 0,
                    const = 4,
                    type = int,
                    help = "usage: TIconvert.py -t 4\n" \
                           "If this is enabled, it indents the code correctly when converting " \
                           "from 8xp to txt.  The number indicates how many spaces each tab " \
                           "should be.")


### Reading input from command line ###
# The parser may return some errors due to erronous user input.
# It quits automatically.
options = parser.parse_args()

### Transforming input into friendlier data ###
Write("Calculating target destination...")
input_path = options.input

if input_path == None:
    parser.parse_args(["--help"])

# Find root directory:
temp = input_path.rfind("\\")
root_path = input_path[0:temp]

# Find target path
if options.target_path == None:
    # Change to default value
    input_name = input_path[temp:]
    if "." in input_name:
        target_name = input_name[0:input_name.index(".")]
    else:
        target_name = input_name
    target_name += "_out.txt"
    target_path = root_path + target_name
else:
    target_path = options.target_path
    if not("." in target_path):
        target_path += ".txt"

# Creating log file
if options.log == "!!!":
    log_path = root_path + "\\TIConvert_log.txt"
    is_logging_enabled = True
elif options.log != None:
    log_path = options.log
    if not("." in target_path):
        target_path += ".txt"
    is_logging_enabled = True

# Quick error check:
DoesFileExist(input_path)
DoesFileExist(target_path)

if is_logging_enabled:
    log_file = open(log_path, "a")
    log_file.write(time.strftime("\n\n%Y-%m-%d %H:%M:%S\n"))   # Timestamp

Write("Input file:  " + input_path)
Write("Target name: " + target_path)
if is_logging_enabled:
    Write("Log file:    " + log_path)



### Preparing buffers ###
Write("Preparing buffers...")
input_file = open(input_path, "rb")  # read-only, binary
try:
    input_buffer = input_file.read() # Copy the entire contents of the file
except:
    QuitProgram("Could not accept input file -- possibly not enough memory?",2)
    
input_file.close()                   # Don't want to leave unclosable files hanging around.
output_buffer = ""

for char in input_buffer:
    output_buffer += NumHexToStr(char)
    

### Validating data ###
errorcount = 0

if output_buffer[0:22] != signature:
    errorcount += 1
    Write("WARNING: The file doesn't start with the correct signature - is it corrupted?", 2)

if (len(output_buffer) / 2 - 57) != ReturnTwoBytes(53, output_buffer):
    errorcount += 1
    Write("WARNING: The filesixe given in the header doesn't match the actual filesize.", 2)

variable_type = ReturnByte(59, output_buffer)
if variable_type in r_AllowedTypes:
    Write("Input file appears to be " + d_VariableTypes[variable_type] + ".")
    if variable_type > 7:
        errorcount != 1
        Write("WARNING: Detokenizing this file may return garbage.", 2)
else:
    errorcount != 1
    Write("WARNING: The file doesn't appear to be a program.", 2)


if errorcount:
    Write("Errors detected, proceeding anyways.", 2)
else:
    Write("No errors detected.")


### Setting up data ###
Write("Parsing preliminary data...")

offset = 74                             # First 74 bytes are header.
data_length = len(output_buffer) // 2 - offset
count = 0
two_byte_flag = False

tab_length = options.tabs

tab = ""
for i in range(tab_length):
    tab += " "

tab_depth = 0

output_file = open(target_path, "w")    # Will overwrite any files with the same name
                                        # Change 2nd arg to "a" for appending.


### Actual detokenizing ###
Write("Detokenizing...")

if options.colon == True:
    output_file.write(":")

for count in range(data_length):
    if count == (data_length - 2):
        # Last two bytes is the checksum, not actual token data
        break

    current_byte = NumHexToStr(ReturnByte(count + offset, output_buffer))

    next_byte = NumHexToStr(ReturnByte(count + offset + 1, output_buffer))
    if count == (data_length - 3):
        # If next byte is part of the checksum, then make it a newline character.
        next_byte = "3F"

    prev_byte = NumHexToStr(ReturnByte(count + offset - 1, output_buffer))
    if count == 0:
        # If this is the first character...
        prev_byte = "3F"

    if current_byte in r_AxePush:
        tab_depth += 1
        if (prev_byte == "D5") and ((current_byte == "CE") or (next_byte == "CE")):
            # Accounting for ReturnIf and Return!If
            tab_depth -= 1
    if next_byte in r_AxePop:
        tab_depth -= 1

    if two_byte_flag == True:
        output_file.write(d_AxeTokensMain[prev_byte][current_byte])
        two_byte_flag = False
    else:
        if current_byte in r_TwoByteTokens:
            two_byte_flag = True
        else:
            output_file.write(d_AxeTokensMain[current_byte])
            if current_byte == "3F":
                if options.colon == True:
                    output_file.write(":")
                if (tab_depth > 0):
                    for i in range(tab_depth):
                        output_file.write(tab)

QuitProgram("Finished")
























